"""Utilitary functions"""

# Import modules
import numpy as np
# from scipy.stats import chi2_contingency

# Function for dummy generation
def _dummygen(x,levs):
    """Generate dummy variables"""
    n_levs = len(levs)

    converted_array = np.empty((x.shape[0],n_levs-1))
    for l in range(n_levs-1):
        converted_array[:,l] = (x == levs[l+1]).astype(int)

    return converted_array

# # Crosstab function
# def _cross(x,y):
#     """Create a cross tab"""
#     tab = []
    
#     for i in np.unique(x):
#         cols = []
        
#         for j in np.unique(y):
#             c = np.count_nonzero((x==i) & (y==j))
#             cols = cols + [c]
        
#         tab = tab + [cols]

#     return np.array(tab)

# Block generation function
def _blockgen(design: np.ndarray, n_blocks: int, ncs: int, reps: int):
    """Blocks generator"""
    # Create array of blocks
    blocks = np.repeat(np.arange(n_blocks)+1,int(ncs/n_blocks))
    # bestcorr = np.inf
    # bestblock = blocks.copy()
    
    for _ in range(reps):
        np.random.shuffle(blocks)
        # blockmat = np.repeat(blocks,int(np.max(design[:,1]))).copy()
        # sumcorr = 0
        
        # for a in range(2,design.shape[1]):
        #     d = design[:,a].copy()
        #     c = _cross(blockmat,d)
        #     corr = chi2_contingency(c)[1]
        #     sumcorr = sumcorr + corr
        
        # if sumcorr < bestcorr:
        #     bestblock = blockmat.copy()
        #     bestcorr = sumcorr

    # return bestblock
    return blocks

# Condition generation function
def _condgen(desname: str, cond: list, names: list, init: bool = False):
    """Conditions generator for the design modules"""    
    # Match variable names with columns in the design matrix
    if init:
        design_columns = [desname + '[i,' + str(i) + ']' for i in range(len(names))]
    else:
        design_columns = [desname + '[:,' + str(i) + ']' for i in range(len(names))]

    # Create new list of conditions
    conditions = []

    for c in cond:
        # Take a copy of the string
        cc = c[:]

        # Replace names by design matrix columns
        for i in range(len(names)):
            cc = cc.replace(names[i],design_columns[i])
        
        # If there is a conditional 'if' statement, then convert it to a logical 'or'
        if 'if' in cc:
            # Split the sting in the 'then' part
            if_part, then_part = cc.split('then')

            # In case there are '&' in the 'if' 'then' part, convert them to '*'
            if_part = if_part.replace('&','*')
            then_part = then_part.replace('&','*')

            # Set the 'logical_not' operator in the 'if' part
            if_part = 'np.logical_not(' + if_part.replace('if','') + ')'

            # Merge if and then parts
            cc = 'np.logical_or(' + if_part + ',' + then_part + ')'
        
        # Finally, append condition to condition list
        conditions.append(cc)
    
    # ...and return the new conditions list
    return conditions

# Generate initial design matrix
def _initdesign(levs: list, ncs: int, cond: list):
    """Generate initial design matrix"""
    # Create and populate the initial design matrix
    desmat = []

    # for k in levs:
    for k in range(len(levs)):
        col = np.array((levs[k] * int(np.ceil(ncs/len(levs[k]))))[:ncs])
        np.random.shuffle(col)
        desmat.append(col)
    
    desmat = np.array(desmat).T

    # Apply conditions if needed
    if cond is not None:
        for i in range(ncs):
            # Check if all conditions are satisfied. If not, do a big enough loop till all conditions are satisfied
            check_all = []

            for c in cond:
                check_all.append(eval(c))

            check_all = np.all(check_all)

            if not check_all:
                for _ in range(10000):
                    # Create a random vector of levels for the row in question
                    for k in range(len(levs)):
                        desmat[i,k] = np.random.choice(levs[k])
                    
                    # Check if conditions are met with the new vector in the design.
                    check_all = []

                    for c in cond:
                        check_all.append(eval(c))

                    check_all = np.all(check_all)


                    # If so, break the loop and go for the next row
                    if check_all:
                        break
            
            # If after the big loop conditions are not met, then raise an error
            assert check_all, 'It is not possible to met all conditions in the initial design matrix.'
    
    # Return the design matrix
    return desmat


"""Optimisation algorithms"""

# Load modules
import datetime
import time
import pandas as pd
import numpy as np
from biogeme.expressions import Expression
from choicedesign.criteria import _derr

# Swapping algorithm function
def _swapalg(
    design: pd.DataFrame, model: Expression,draws: int,
    init_perf: float, cond: list,
    iter_lim: float, noimprov_lim: float,time_lim: float):
    """Random swapping algorithm

    It optimises an experimental design using a variation of the random swapping 
    algorithm [1].

    References
    ----------
    [1] Quan, W., Rose, J. M., Collins, A. T., & Bliemer, M. C. (2011). A comparison 
    of algorithms for generating efficient choice experiments.
    """
    # Lock design matrix and names
    names = design.columns
    desmat = design.to_numpy()

    # Start stopwatch
    t0 = time.time()
    t1 = time.time()

    difftime = 0

    # Initialize algorithm parameters
    i = np.random.choice(np.arange(design.shape[1]))
    t = 0
    ni = 0
    iterperf = init_perf
    newperf = init_perf

    # Start algorithm
    while True:
        
        # Iteration No.
        t = t+1
        
        # If one stopping criterion is satisfied, break!
        if ni >= noimprov_lim or t >= iter_lim or (difftime)/60 >= time_lim:
            break
        
        # Take a random swap
        pairswap = np.random.choice(desmat.shape[0],2,replace=False)
        
        # Check if attribute levels differ
        check_difflevels = desmat[pairswap[0],i] != desmat[pairswap[1],i]

        # If attribute levels differ, do the swap and check for conditions (if defined)
        if check_difflevels:
            swapdes = desmat.copy()
            swapdes[pairswap[0],i] = desmat[pairswap[1],i]
            swapdes[pairswap[1],i] = desmat[pairswap[0],i]
        
            # Check if conditions are satisfied after a swap
            check_all = [True]
            
            # If conditions are defined, this section will check that are satisfied, and rewrite 'check_satisfied_conds' if neccesary
            if cond is not None:
                for c in cond:
                    check_all.append(np.all(eval(c)))

                check_all = np.all(check_all)
            
            # If all conditions are satisfied, compute D-error
            if check_all:
                newperf = _derr(pd.DataFrame(swapdes,columns=names),model,draws)

        # ...else if they do not differ, keep the D-error
        else:
            newperf = iterperf.copy()
            
        # If the swap made an improvement, keep the design and update progress bar
        improved = newperf < iterperf

        if improved:
            desmat = swapdes.copy()
            iterperf = newperf.copy()
            ni = 0
            
            # Update progress bar
            print('Optimizing / ' + 'Elapsed: ' + str(datetime.timedelta(seconds=difftime))[:7] + ' / D-error: ' + str(round(iterperf,6)),end='\r')
        
        # ...else, pass to a random attribute and increment the 'no improvement' counter by 1.
        else:
            i = np.random.choice(np.arange(design.shape[1]))
            ni = ni+1
        
        # Update progress bar each second
        if (difftime)%1 < 0.1:
            print('Optimizing / ' + 'Elapsed: ' + str(datetime.timedelta(seconds=difftime))[:7] + ' / D-error: ' + str(round(iterperf,6)),end='\r',flush=True)
        
        t1 = time.time()
        difftime = t1-t0
    
    # Return optimal design plus efficiency
    return pd.DataFrame(desmat,columns=names), iterperf, t, difftime


"""Optimisation criteria"""

# Import modules
import numpy as np
import pandas as pd
from biogeme.expressions import Expression
from biogeme.database import Database
from choicedesign.utils import _dummygen

# D-error function
def _derr(design: pd.DataFrame, model: Expression,ndraws):
    """D-error of an information matrix"""
    # Get information matrix
    _, _, im, _ = model.getValueAndDerivatives(database=Database('design',design),numberOfDraws=ndraws,aggregation=True,prepareIds=True,bhhh=False)

    # Calculate D-error
    if np.linalg.det(im) != 0:
        vce = np.linalg.solve(im,np.eye(im.shape[0]))

        # if asc is not None:
        #     vce = vce[len(asc):,len(asc):]

        detvce = np.linalg.det(vce)

        if detvce > 0:
            dr = detvce**(1/vce.shape[0])
        else:
            dr = np.inf
    else:
        dr = np.inf

    return dr

# Utility balance
def _utility_balance(design: pd.DataFrame, model: list,ndraws):
    """Utility balance ratio"""
    # Calculate choice probabilities
    p = []

    for j in range(len(model)):
        p.append(model[j].getValueAndDerivatives(database=Database('design',design),numberOfDraws=ndraws,aggregation=False,prepareIds=True,bhhh=False)[0])

    p = np.array(p)

    # Calculate utility balance ratio
    B = p/(1/len(p))
    B = np.prod(B,axis=0)*100
    B = np.sum(B)/len(design)

    return B

# # Information matrix of MNL model
# def _imat_mnl(design: np.ndarray, levs: list, coding: list, pars: list, natts: int, nalts: int, ncs: int, optout: bool, asc: dict):
#     """Information matrix of the MNL model"""
#     # Populate the coded matrix
#     coded_matrix = []

#     for k in range(len(coding)):
#         if coding[k] == 'dummy':
#             dd = _dummygen(design[:,k],levs[k])
#             coded_matrix.append(dd)
#         else:
#             coded_matrix.append(design[:,k][:,np.newaxis])
    
#     coded_matrix = np.concatenate(coded_matrix,axis=1)

#     # If optout alternative is present, add to coded matrix
#     if optout:
#         len_asc = (len(asc) if asc is not None else 0)
#         coded_matrix = np.c_[coded_matrix,np.zeros((ncs,int(len(pars)/(nalts+optout))-len_asc))]
        
#     # Calculate utility of each alternative
#     v = coded_matrix * pars
#     v.shape = (ncs, nalts+optout, int(len(pars)/(nalts+optout)))
#     v = v.sum(axis=2)

#     # Calculate Probability
#     ev = np.exp(v)
#     sev = ev.sum(axis=1,keepdims=True)
#     p = (ev/sev).flatten()

#     # Calculate Information Matrix
#     coded_matrix.shape = (ncs * (nalts+optout), int(len(pars)/(nalts+optout)))
    
#     # Calculate Information Matrix
#     ia = np.diag(p) @ coded_matrix
#     ia = ia.T @ coded_matrix
    
#     ib = np.diag(p) @ coded_matrix
#     ib.shape = (ncs,nalts+optout,int(len(pars)/(nalts+optout)))
#     ib = np.sum(ib,axis=1)
#     ib = ib.T @ ib
    
#     im = ia - ib

#     # Return information matrix
#     return im

# # Utility balance function
# def _utility_balance_mnl(design: np.ndarray, levs: list, coding: list, pars: list, natts: int, nalts: int, ncs: int, optout: bool, asc: dict):
#     """Utility balance ratio of the MNL model"""
#     # Populate the coded matrix
#     coded_matrix = []

#     for k in range(len(coding)):
#         if coding[k] == 'dummy':
#             dd = _dummygen(design[:,k],levs[k])
#             coded_matrix.append(dd)
#         else:
#             coded_matrix.append(design[:,k][:,np.newaxis])
    
#     coded_matrix = np.concatenate(coded_matrix,axis=1)

#     # If optout alternative is present, add to coded matrix
#     if optout:
#         len_asc = (len(asc) if asc is not None else 0)
#         coded_matrix = np.c_[coded_matrix,np.zeros((ncs,int(len(pars)/(nalts+optout))-len_asc))]
        
#     # Calculate utility of each alternative
#     v = coded_matrix * pars
#     v.shape = (ncs, nalts+optout, int(len(pars)/(nalts+optout)))
#     v = v.sum(axis=2)

#     # Calculate Probability
#     ev = np.exp(v)
#     sev = ev.sum(axis=1,keepdims=True)
#     p = (ev/sev)

#     # Calculate utility balance
#     B = (p/(1/(nalts+optout))).copy()
#     B = np.prod(B,axis=1)*100
#     B = np.sum(B)/ncs

#     return B


"""Modules to construct efficient designs"""

# Import modules
from re import L
from unicodedata import name
import pandas as pd
import numpy as np
import datetime
from biogeme.expressions import Expression, MonteCarlo, log
from biogeme.models import loglogit, logit

from choicedesign.algorithms import _swapalg
from choicedesign.criteria import _derr, _utility_balance
from choicedesign.utils import _blockgen, _condgen, _initdesign

# Efficient design
class EffDesign:
    """RUM-consistent discrete choice experiment design class

    This class allows to create an efficient design for a discrete choice 
    experiment [1].

    Parameters
    ----------
    alts : list
        List of alternative names, to generate the design matrix
    ncs : int
        Number of choice situations.
    atts_list : list[dict]
        List of attributes of the design. Each element is a dictionary that 
        contains the following keys:
            
            - `name`: the attribute name
            - `levels`: a list of attribute levels
            - `avail`: a list that details whether the attribute is part of 
            a specific alternative. Each element is one of the alternative 
            names of the parameter `alts`.
    """

    # Init method
    def __init__(self,atts_list: list, alts: list, ncs: int):

        # Define scalars
        self.N = ncs
        self.J = len(alts)
        self.K = len(atts_list)
 
        # Set names and levels
        self.names = []
        self.levs = []

        for j in alts:
            for k in atts_list:
                if j in k['avail']:
                    self.names.append(j + '_' + k['name'])
                    self.levs.append(k['levels'])

    # Generate initial design matrix
    def gen_initdesign(self,cond: list = None, seed: bool = None):
        """Generate initial design matrix

        It generates the initial design matrix. The user can define a set of
        conditions that must be satisfied.

        Parameters
        -------
        cond : list[str], optional
            List of conditions that the final design must hold. Each element 
            is a string that contains a single condition. Conditions 
            can be of the form of binary relations (e.g., `X > Y` where `X` 
            and `Y` are attributes of a specific alternative) or conditional 
            relations (e.g., `if X > a then Y < b` where `a` and `b` are values).
            Users can specify multiple conditions when the operator `if` is defined, 
            separated by the operator `&`, by default None
        seed : bool, None
            Random seed, by default None

        Returns
        -------
        init_design : pandas.DataFrame
            A Pandas DataFrame with the initial design matrix.
        """

        # Generate conditions if defined
        if cond is not None:
            self.initconds = _condgen('desmat',cond,self.names,init=True)
            self.algconds = _condgen('swapdes',cond,self.names,init=False)
        else:
            self.initconds = None
            self.algconds = None

        # Set random seed if defined
        if seed is not None:
            np.random.seed(seed)

        # Generate initial design matrix
        init_design = _initdesign(levs=self.levs,ncs=self.N,cond=self.initconds)

        return pd.DataFrame(init_design,columns=self.names)

    # Optimise
    def optimise(self, init_design: pd.DataFrame, V: dict, model: str = 'mnl', draws: int = 1000, n_blocks: int = None, iter_lim: int = None, noimprov_lim: int = None, time_lim: int = None, seed: int = None, verbose: bool = False):
        """Create D-efficient RUM design

        Starts the optimisation of the design using a random swapping 
        algorithm and the attributes and prior parameters specified in 
        the `RUMDesign` object. Allows for conditions, blocking and 
        user-defined stopping criteria.

        Parameters
        ----------
        initial_design : pandas.DataFrame
            The initial design matrix as a Pandas DataFrame
        V : dict
            A dictionary with the utility function, using the same syntax as in Biogeme
        model : str
            The base model for the efficient design, by default 'mnl'
        draws : int, optional
            Number of draws for the Monte Carlo integration, by default 1000
        n_blocks : int, optional
            Number of blocks of the final design. Must be a multiple of the number of 
            choice situations, by default None
        iter_lim : int, optional
            Number of iterations before the algorithm stops, by default None
        noimprov_lim : int, optional
            Number of iterations without improvement before the algorithm 
            stops, by default None
        time_lim : int, optional
            Time (in minutes) before the algorithm stops, by default None
        seed : int, optional
            Random seed, by default None
        verbose : bool, optional
            Whether status messages and progress are shown, by default False

        Returns
        -------
        optimal_design : pandas.DataFrame
            The final (optimal) design
        init_perf : float
            D-error of the initial design
        final_perf : float
            D-error of the final design 
        final_iter : int
            Total number of iterations
        ubalance_ratio : float
            Utility balance ratio
        """
        # Set random seed if defined
        if seed is not None:
            np.random.seed(seed)

        # Set stopping criteria if defined
        if iter_lim is None:
            iter_lim = np.inf
        
        if noimprov_lim is None:
            noimprov_lim = np.inf
            
        if time_lim is None:
            time_lim = np.inf

        ############################################################
        ########## Step 1: Set initial design performance ##########
        ############################################################

        if verbose:
            print('Evaluating initial design')

        desmat = init_design

        models_ubalance = []
        av = dict()
        if model == 'mnl':
            for k, _ in V.items():
                av[k] = 1
            model_object = loglogit(V,av,1)

            for k, _ in V.items():
                models_ubalance.append(logit(V,av,k))
        elif model == 'mnl_bayesian':
            for k, _ in V.items():
                av[k] = 1
            model_object = log(MonteCarlo(logit(V,av,1)))

            for k, _ in V.items():
                models_ubalance.append(MonteCarlo(logit(V,av,k)))
        else:
            raise ValueError("""Model name must be either 'mnl' or 'mnl_bayesian'""")

        init_perf = _derr(desmat,model_object,draws)

        ############################################################
        ############## Step 2: Initialize algorighm ################
        ############################################################

        # Execute Swapping algorithm
        optimal_design, final_perf, final_iter, elapsed_time = _swapalg(
            desmat,model_object,draws,init_perf,self.algconds,iter_lim,noimprov_lim,time_lim)

        # Compute utility balance ratio
        ubalance_ratio = _utility_balance(pd.DataFrame(optimal_design,columns=self.names),models_ubalance,draws)

        ############################################################
        ############## Step 3: Arange final design #################
        ############################################################

        # Add CS column
        optimal_design = np.c_[np.arange(self.N)+1,optimal_design]

        # Generate blocks
        if n_blocks is not None:
            if verbose:
                print('\nGenerating ' + str(n_blocks) + ' blocks...')
            blocksrow = _blockgen(optimal_design,n_blocks,self.N,1000)
            optimal_design = np.c_[optimal_design,blocksrow]

        # Create Pandas DataFrame
        if n_blocks is not None:
            optimal_design = pd.DataFrame(optimal_design,columns=['CS'] + self.names + ['Block'])
        else:
            optimal_design = pd.DataFrame(optimal_design,columns=['CS'] + self.names)

        # Return a summary if verbose is True
        if verbose:
            print('Optimization complete')
            print('Elapsed time: ' + str(datetime.timedelta(seconds=elapsed_time))[:7])
            print('D-error of initial design: ',round(init_perf,6))
            print('D-error of last stored design: ',round(final_perf,6))
            print('Utility Balance ratio: ',round(ubalance_ratio,2),'%')
            print('Algorithm iterations: ',final_iter)
            print('')
        
        # Return the optimal design
        return optimal_design, init_perf, final_perf, final_iter, ubalance_ratio

    # Evaluate
    def evaluate(self, design: pd.DataFrame, V: dict, model: str = 'mnl', draws: int = 1000):
        """Evaluate design

        Evaluates a design stored in a Pandas data frame

        Parameters
        ----------
        design : pd.DataFrame
            Design to evaluate
        V : dict
            A dictionary with the utility function, using the same syntax as in Biogeme
        model : str
            The base model for the efficient design, by default 'mnl'

        Returns
        -------
        perf : float
            The D-error of the design
        ubalance_ratio : float
            Utility balance ratio
        """
        # Drop CS column and Block (if present) from pandas dataframe
        desmat = design.drop('CS',axis=1)

        if 'Block' in desmat.columns:
            desmat = desmat.drop('Block',axis=1)

        models_ubalance = []
        av = dict()
        if model == 'mnl':
            for k, _ in V.items():
                av[k] = 1
            model_object = loglogit(V,av,1)

            for k, _ in V.items():
                models_ubalance.append(logit(V,av,k))
        elif model == 'mnl_bayesian':
            for k, _ in V.items():
                av[k] = 1
            model_object = log(MonteCarlo(logit(V,av,1)))

            for k, _ in V.items():
                models_ubalance.append(MonteCarlo(logit(V,av,k)))
        else:
            raise ValueError("""Model name must be either 'mnl' or 'mnl_bayesian'""")
        
        # Evaluate the performance and utility balance of the design
        perf = _derr(desmat,model_object,draws)
        ubalance_ratio = _utility_balance(desmat,models_ubalance,draws)

        # Return performance and utility balance
        return perf, ubalance_ratio


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['choicedesign']

package_data = \
{'': ['*']}

install_requires = \
['biogeme>=3.2,<4.0', 'numpy>=1.23.4,<2.0.0', 'pandas>=1.5.1,<2.0.0']

setup_kwargs = {
    'name': 'choicedesign',
    'version': '0.1.5',
    'description': 'Experimental designs for discrete choice models.',
    'long_description': '# ChoiceDesign\n\n**ChoiceDesign** is a Python package tool to construct D-efficient designs for Discrete Choice Experiments. ChoiceDesign combines enough flexibility to construct from simple 2-alternative designs with few attributes, to more complex settings that may involve conditions between attributes. ChoiceDesign is a revamped version of [EDT](https://github.com/ighdez/EDT), a project I created some years ago for the same purpose. ChoiceDesign includes improvements over EDT such as class-based syntax, coding improvements, better documentation and making this package available to install via `pip`.\n\n**NEW:** ChoiceDesign now integrates [Biogeme](https://biogeme.epfl.ch/), which allows you to customise the utility functions.\n\n## Installation\n\nChoiceDesign is available to install via the regular syntax of `pip`:\n\n* ``python3 -m pip install choicedesign``\n\n## Features\n\nThe main features of ChoiceDesign are:\n\n* D-efficient designs based on a random swapping algorithm\n* **Customisable utility functions** (powered by [Biogeme](https://biogeme.epfl.ch/))\n* Bayesian priors *(experimental)*\n* Designs with conditions over different attribute levels\n* Designs with blocks.\n* Multiple stopping criteria (Fixed number of iterations, iterations without improvement or fixed time).\n\n## Examples\n\nI provide some Jupyter notebooks that illustrate the use of ChoiceDesign in the `examples/` folder of this repo.\n\n## How to contribute?\nAny contributions to ChoiceDesign are welcome via this Git, or to my email joseignaciohernandezh at gmail dot com.\n\n## Disclaimer\n\nThis software is provided for free and as it is, say with **no warranty**, and neither me nor my current institution is liable of any consequence of the use of it. In any case, integrity checks have been performed by comparing results with alternative software.\n\n## References\n* Bierlaire, M. (2003). BIOGEME: A free package for the estimation of discrete choice models. In *Swiss transport research conference*.\n* Kuhfeld, W. F. (2005). Experimental design, efficiency, coding, and choice designs. *Marketing research methods in SAS: Experimental design, choice, conjoint, and graphical techniques*, 47-97.\n* Quan, W., Rose, J. M., Collins, A. T., & Bliemer, M. C. (2011). A comparison of algorithms for generating efficient choice experiments.',
    'author': 'José Ignacio Hernández',
    'author_email': 'joseignaciohernandezh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
